@@ -23,6 +23,7 @@
 #include "bson.h"
 #include "bson-config.h"
 #include "bson-json.h"
+#include "bson-json-private.h"
 #include "bson-iso8601-private.h"
 
 #include "common-b64-private.h"
@@ -102,7 +103,7 @@ static const char *read_state_names[] = {FOREACH_READ_STATE (GENERATE_STRING)};
    BS (DECIMAL128)                                                   \
    BS (DBPOINTER)                                                    \
    BS (SYMBOL)                                                       \
-   BS (DBREF)
+   BS (UUID)
 
 typedef enum {
    FOREACH_BSON_STATE (BSON_STATE_ENUM)
@@ -129,8 +130,6 @@ typedef enum {
 typedef struct {
    int i;
    bson_json_frame_type_t type;
-   bool has_ref;
-   bool has_id;
    bson_t bson;
 } bson_json_stack_frame_t;
 
@@ -277,8 +276,6 @@ _noop (void)
 #define FRAME_TYPE_HAS_BSON(_type) \
    ((_type) == BSON_JSON_FRAME_SCOPE || (_type) == BSON_JSON_FRAME_DBPOINTER)
 #define STACK_HAS_BSON FRAME_TYPE_HAS_BSON (STACK_FRAME_TYPE)
-#define STACK_HAS_REF STACK_ELE (0, has_ref)
-#define STACK_HAS_ID STACK_ELE (0, has_id)
 #define STACK_PUSH(frame_type)                       \
    do {                                              \
       if (bson->n >= (STACK_MAX - 1)) {              \
@@ -307,8 +304,6 @@ _noop (void)
 #define STACK_PUSH_DOC(statement)       \
    do {                                 \
       STACK_PUSH (BSON_JSON_FRAME_DOC); \
-      STACK_HAS_REF = false;            \
-      STACK_HAS_ID = false;             \
       if (bson->n != 0) {               \
          statement;                     \
       }                                 \
@@ -375,8 +370,9 @@ _noop (void)
                                  read_state_names[bson->read_state]);          \
       return;                                                                  \
    }
-#define HANDLE_OPTION(_key, _type, _state)                                     \
-   (len == strlen (_key) && strncmp ((const char *) val, (_key), len) == 0)    \
+#define HANDLE_OPTION(_selection_statement, _key, _type, _state)               \
+   _selection_statement (len == strlen (_key) &&                               \
+                         strncmp ((const char *) val, (_key), len) == 0)       \
    {                                                                           \
       if (bson->bson_type && bson->bson_type != (_type)) {                     \
          _bson_json_read_set_error (reader,                                    \
@@ -392,6 +388,24 @@ _noop (void)
    }
 
 
+bson_json_opts_t *
+bson_json_opts_new (bson_json_mode_t mode, int32_t max_len)
+{
+   bson_json_opts_t *opts;
+
+   opts = (bson_json_opts_t *) bson_malloc (sizeof *opts);
+   opts->mode = mode;
+   opts->max_len = max_len;
+
+   return opts;
+}
+
+void
+bson_json_opts_destroy (bson_json_opts_t *opts)
+{
+   bson_free (opts);
+}
+
 static void
 _bson_json_read_set_error (bson_json_reader_t *reader, const char *fmt, ...)
    BSON_GNUC_PRINTF (2, 3);
@@ -686,12 +700,12 @@ _bson_json_read_integer (bson_json_reader_t *reader, uint64_t val, int64_t sign)
       case BSON_JSON_LF_OID:
       case BSON_JSON_LF_BINARY:
       case BSON_JSON_LF_TYPE:
+      case BSON_JSON_LF_UUID:
       case BSON_JSON_LF_UNDEFINED:
       case BSON_JSON_LF_DOUBLE:
       case BSON_JSON_LF_DECIMAL128:
       case BSON_JSON_LF_DBPOINTER:
       case BSON_JSON_LF_SYMBOL:
-      case BSON_JSON_LF_DBREF:
       default:
          _bson_json_read_set_error (reader,
                                     "Unexpected integer %s%" PRIu64
@@ -813,8 +827,29 @@ _bson_json_read_int64_or_set_error (bson_json_reader_t *reader, /* IN */
    return true;
 }
 
+static bool
+_unhexlify_uuid (const char *uuid, uint8_t *out, size_t max)
+{
+   unsigned int byte;
+   int x = 0;
+   int i = 0;
 
-/* parse a value for "base64", "subType" or legacy "$binary" or "$type" */
+   BSON_ASSERT (strlen (uuid) == 32);
+
+   while (SSCANF (&uuid[i], "%2x", &byte) == 1) {
+      if (x >= max) {
+         return false;
+      }
+
+      out[x++] = (uint8_t) byte;
+      i += 2;
+   }
+
+   return i == 32;
+}
+
+/* parse a value for "base64", "subType", legacy "$binary" or "$type", or
+ * "$uuid" */
 static void
 _bson_json_parse_binary_elem (bson_json_reader_t *reader,
                               const char *val_w_null,
@@ -831,7 +866,7 @@ _bson_json_parse_binary_elem (bson_json_reader_t *reader,
 
    if (bs == BSON_JSON_LF_BINARY) {
       data->binary.has_binary = true;
-      binary_len = bson_b64_pton (val_w_null, NULL, 0);
+      binary_len = mcommon_b64_pton (val_w_null, NULL, 0);
       if (binary_len < 0) {
          _bson_json_read_set_error (
             reader,
@@ -840,9 +875,9 @@ _bson_json_parse_binary_elem (bson_json_reader_t *reader,
       }
 
       _bson_json_buf_ensure (&bson->bson_type_buf[0], (size_t) binary_len + 1);
-      if (bson_b64_pton (val_w_null,
-                         bson->bson_type_buf[0].buf,
-                         (size_t) binary_len + 1) < 0) {
+      if (mcommon_b64_pton (val_w_null,
+                            bson->bson_type_buf[0].buf,
+                            (size_t) binary_len + 1) < 0) {
          _bson_json_read_set_error (
             reader,
             "Invalid input string \"%s\", looking for base64-encoded binary",
@@ -874,9 +909,95 @@ _bson_json_parse_binary_elem (bson_json_reader_t *reader,
                               (int) vlen);
          }
       }
+   } else if (bs == BSON_JSON_LF_UUID) {
+      int nread = 0;
+      char uuid[33];
+
+      data->binary.has_binary = true;
+      data->binary.has_subtype = true;
+      data->binary.type = BSON_SUBTYPE_UUID;
+
+      /* Validate the UUID and extract relevant portions */
+      /* We can't use %x here as it allows +, -, and 0x prefixes */
+#ifdef _MSC_VER
+      SSCANF (val_w_null,
+              "%8c-%4c-%4c-%4c-%12c%n",
+              &uuid[0],
+              8,
+              &uuid[8],
+              4,
+              &uuid[12],
+              4,
+              &uuid[16],
+              4,
+              &uuid[20],
+              12,
+              &nread);
+#else
+      SSCANF (val_w_null,
+              "%8c-%4c-%4c-%4c-%12c%n",
+              &uuid[0],
+              &uuid[8],
+              &uuid[12],
+              &uuid[16],
+              &uuid[20],
+              &nread);
+#endif
+
+      uuid[32] = '\0';
+
+      if (nread != 36 || val_w_null[nread] != '\0') {
+         _bson_json_read_set_error (reader,
+                                    "Invalid input string \"%s\", looking for "
+                                    "a dash-separated UUID string",
+                                    val_w_null);
+
+         return;
+      }
+
+      binary_len = 16;
+      _bson_json_buf_ensure (&bson->bson_type_buf[0], (size_t) binary_len + 1);
+
+      if (!_unhexlify_uuid (
+             &uuid[0], bson->bson_type_buf[0].buf, (size_t) binary_len)) {
+         _bson_json_read_set_error (reader,
+                                    "Invalid input string \"%s\", looking for "
+                                    "a dash-separated UUID string",
+                                    val_w_null);
+      }
+
+      bson->bson_type_buf[0].len = (size_t) binary_len;
    }
 }
 
+static bool
+_bson_json_allow_embedded_nulls (bson_json_reader_t const *reader)
+{
+   const bson_json_read_state_t read_state = reader->bson.read_state;
+   const bson_json_read_bson_state_t bson_state = reader->bson.bson_state;
+
+   if (read_state == BSON_JSON_IN_BSON_TYPE_REGEX_VALUES) {
+      if (bson_state == BSON_JSON_LF_REGULAR_EXPRESSION_PATTERN ||
+          bson_state == BSON_JSON_LF_REGULAR_EXPRESSION_OPTIONS) {
+         /* Prohibit embedded NULL bytes for canonical extended regex:
+          * { $regularExpression: { pattern: "pattern", options: "options" } }
+          */
+         return false;
+      }
+   }
+
+   if (read_state == BSON_JSON_IN_BSON_TYPE) {
+      if (bson_state == BSON_JSON_LF_REGEX ||
+          bson_state == BSON_JSON_LF_OPTIONS) {
+         /* Prohibit embedded NULL bytes for legacy regex:
+          * { $regex: "pattern", $options: "options" } */
+         return false;
+      }
+   }
+
+   /* Embedded nulls are okay in any other context */
+   return true;
+}
 
 static void
 _bson_json_read_string (bson_json_reader_t *reader, /* IN */
@@ -885,13 +1006,14 @@ _bson_json_read_string (bson_json_reader_t *reader, /* IN */
 {
    bson_json_read_state_t rs;
    bson_json_read_bson_state_t bs;
+   const bool allow_null = _bson_json_allow_embedded_nulls (reader);
 
    BASIC_CB_PREAMBLE;
 
    rs = bson->read_state;
    bs = bson->bson_state;
 
-   if (!bson_utf8_validate ((const char *) val, vlen, true /*allow null*/)) {
+   if (!bson_utf8_validate ((const char *) val, vlen, allow_null)) {
       _bson_json_read_corrupt (reader, "invalid bytes in UTF8 string");
       return;
    }
@@ -947,6 +1069,8 @@ _bson_json_read_string (bson_json_reader_t *reader, /* IN */
       case BSON_JSON_LF_BINARY:
       case BSON_JSON_LF_TYPE:
          bson->bson_type_data.binary.is_legacy = true;
+         /* FALL THROUGH */
+      case BSON_JSON_LF_UUID:
          _bson_json_parse_binary_elem (reader, val_w_null, vlen);
          break;
       case BSON_JSON_LF_INT32: {
@@ -1020,12 +1144,6 @@ _bson_json_read_string (bson_json_reader_t *reader, /* IN */
          bson_append_symbol (
             STACK_BSON_CHILD, key, (int) len, (const char *) val, (int) vlen);
          break;
-      case BSON_JSON_LF_DBREF:
-         /* the "$ref" of a {$ref: "...", $id: ... }, append normally */
-         bson_append_utf8 (
-            STACK_BSON_CHILD, key, (int) len, (const char *) val, (int) vlen);
-         bson->read_state = BSON_JSON_REGULAR;
-         break;
       case BSON_JSON_LF_SCOPE:
       case BSON_JSON_LF_TIMESTAMP_T:
       case BSON_JSON_LF_TIMESTAMP_I:
@@ -1056,11 +1174,14 @@ _bson_json_read_start_map (bson_json_reader_t *reader) /* IN */
    BASIC_CB_PREAMBLE;
 
    if (bson->read_state == BSON_JSON_IN_BSON_TYPE) {
-      if (bson->bson_state == BSON_JSON_LF_DATE) {
+      switch (bson->bson_state) {
+      case BSON_JSON_LF_DATE:
          bson->read_state = BSON_JSON_IN_BSON_TYPE_DATE_NUMBERLONG;
-      } else if (bson->bson_state == BSON_JSON_LF_BINARY) {
+         break;
+      case BSON_JSON_LF_BINARY:
          bson->read_state = BSON_JSON_IN_BSON_TYPE_BINARY_VALUES;
-      } else if (bson->bson_state == BSON_JSON_LF_TYPE) {
+         break;
+      case BSON_JSON_LF_TYPE:
          /* special case, we started parsing {$type: {$numberInt: "2"}} and we
           * expected a legacy Binary format. now we see the second "{", so
           * backtrack and parse $type query operator. */
@@ -1068,6 +1189,44 @@ _bson_json_read_start_map (bson_json_reader_t *reader) /* IN */
          STACK_PUSH_DOC (bson_append_document_begin (
             STACK_BSON_PARENT, key, len, STACK_BSON_CHILD));
          _bson_json_save_map_key (bson, (const uint8_t *) "$type", 5);
+         break;
+      case BSON_JSON_LF_CODE:
+      case BSON_JSON_LF_DECIMAL128:
+      case BSON_JSON_LF_DOUBLE:
+      case BSON_JSON_LF_INT32:
+      case BSON_JSON_LF_INT64:
+      case BSON_JSON_LF_MAXKEY:
+      case BSON_JSON_LF_MINKEY:
+      case BSON_JSON_LF_OID:
+      case BSON_JSON_LF_OPTIONS:
+      case BSON_JSON_LF_REGEX:
+      /**
+       * NOTE: A read_state of BSON_JSON_IN_BSON_TYPE is used when "$regex" is
+       * found, but BSON_JSON_IN_BSON_TYPE_REGEX_STARTMAP is used for
+       * "$regularExpression", which will instead go to a below 'if else' branch
+       * instead of this switch statement. They're both called "regex" in their
+       * respective enumerators, but they behave differently when parsing.
+       */
+      // fallthrough
+      case BSON_JSON_LF_REGULAR_EXPRESSION_OPTIONS:
+      case BSON_JSON_LF_REGULAR_EXPRESSION_PATTERN:
+      case BSON_JSON_LF_SYMBOL:
+      case BSON_JSON_LF_UNDEFINED:
+      case BSON_JSON_LF_UUID:
+         // These special keys do not expect objects as their values. Fail.
+         _bson_json_read_set_error (
+            reader,
+            "Unexpected nested object value for \"%s\" key",
+            reader->bson.unescaped.buf);
+         break;
+      case BSON_JSON_LF_DBPOINTER:
+      case BSON_JSON_LF_SCOPE:
+      case BSON_JSON_LF_TIMESTAMP_I:
+      case BSON_JSON_LF_TIMESTAMP_T:
+      default:
+         // These special LF keys aren't handled with BSON_JSON_IN_BSON_TYPE
+         BSON_UNREACHABLE (
+            "These LF values are handled with a different read_state");
       }
    } else if (bson->read_state == BSON_JSON_IN_BSON_TYPE_TIMESTAMP_STARTMAP) {
       bson->read_state = BSON_JSON_IN_BSON_TYPE_TIMESTAMP_VALUES;
@@ -1103,7 +1262,7 @@ _is_known_key (const char *key, size_t len)
           IS_KEY ("$numberDouble") || IS_KEY ("$numberDecimal") ||
           IS_KEY ("$numberInt") || IS_KEY ("$numberLong") ||
           IS_KEY ("$numberDouble") || IS_KEY ("$numberDecimal") ||
-          IS_KEY ("$dbPointer") || IS_KEY ("$symbol"));
+          IS_KEY ("$dbPointer") || IS_KEY ("$symbol") || IS_KEY ("$uuid"));
 
 #undef IS_KEY
 
@@ -1180,7 +1339,7 @@ _bson_json_read_map_key (bson_json_reader_t *reader, /* IN */
 {
    bson_json_reader_bson_t *bson = &reader->bson;
 
-   if (!bson_utf8_validate ((const char *) val, len, true /* allow null */)) {
+   if (!bson_utf8_validate ((const char *) val, len, false /* allow null */)) {
       _bson_json_read_corrupt (reader, "invalid bytes in UTF8 string");
       return;
    }
@@ -1211,106 +1370,96 @@ _bson_json_read_map_key (bson_json_reader_t *reader, /* IN */
    }
 
    if (bson->read_state == BSON_JSON_IN_BSON_TYPE) {
-      if
-         HANDLE_OPTION ("$regex", BSON_TYPE_REGEX, BSON_JSON_LF_REGEX)
-      else if
-         HANDLE_OPTION ("$options", BSON_TYPE_REGEX, BSON_JSON_LF_OPTIONS)
-      else if
-         HANDLE_OPTION ("$oid", BSON_TYPE_OID, BSON_JSON_LF_OID)
-      else if
-         HANDLE_OPTION ("$binary", BSON_TYPE_BINARY, BSON_JSON_LF_BINARY)
-      else if
-         HANDLE_OPTION ("$type", BSON_TYPE_BINARY, BSON_JSON_LF_TYPE)
-      else if
-         HANDLE_OPTION ("$date", BSON_TYPE_DATE_TIME, BSON_JSON_LF_DATE)
-      else if
-         HANDLE_OPTION (
-            "$undefined", BSON_TYPE_UNDEFINED, BSON_JSON_LF_UNDEFINED)
-      else if
-         HANDLE_OPTION ("$minKey", BSON_TYPE_MINKEY, BSON_JSON_LF_MINKEY)
-      else if
-         HANDLE_OPTION ("$maxKey", BSON_TYPE_MAXKEY, BSON_JSON_LF_MAXKEY)
-      else if
-         HANDLE_OPTION ("$numberInt", BSON_TYPE_INT32, BSON_JSON_LF_INT32)
-      else if
-         HANDLE_OPTION ("$numberLong", BSON_TYPE_INT64, BSON_JSON_LF_INT64)
-      else if
-         HANDLE_OPTION ("$numberDouble", BSON_TYPE_DOUBLE, BSON_JSON_LF_DOUBLE)
-      else if
-         HANDLE_OPTION ("$symbol", BSON_TYPE_SYMBOL, BSON_JSON_LF_SYMBOL)
-      else if
-         HANDLE_OPTION (
-            "$numberDecimal", BSON_TYPE_DECIMAL128, BSON_JSON_LF_DECIMAL128)
-      else if (!strcmp ("$timestamp", (const char *) val)) {
+      HANDLE_OPTION (if, "$regex", BSON_TYPE_REGEX, BSON_JSON_LF_REGEX)
+      HANDLE_OPTION (else if, "$options", BSON_TYPE_REGEX, BSON_JSON_LF_OPTIONS)
+      HANDLE_OPTION (else if, "$oid", BSON_TYPE_OID, BSON_JSON_LF_OID)
+      HANDLE_OPTION (else if, "$binary", BSON_TYPE_BINARY, BSON_JSON_LF_BINARY)
+      HANDLE_OPTION (else if, "$type", BSON_TYPE_BINARY, BSON_JSON_LF_TYPE)
+      HANDLE_OPTION (else if, "$uuid", BSON_TYPE_BINARY, BSON_JSON_LF_UUID)
+      HANDLE_OPTION (else if, "$date", BSON_TYPE_DATE_TIME, BSON_JSON_LF_DATE)
+      HANDLE_OPTION (
+         else if, "$undefined", BSON_TYPE_UNDEFINED, BSON_JSON_LF_UNDEFINED)
+      HANDLE_OPTION (else if, "$minKey", BSON_TYPE_MINKEY, BSON_JSON_LF_MINKEY)
+      HANDLE_OPTION (else if, "$maxKey", BSON_TYPE_MAXKEY, BSON_JSON_LF_MAXKEY)
+      HANDLE_OPTION (else if, "$numberInt", BSON_TYPE_INT32, BSON_JSON_LF_INT32)
+      HANDLE_OPTION (
+         else if, "$numberLong", BSON_TYPE_INT64, BSON_JSON_LF_INT64)
+      HANDLE_OPTION (
+         else if, "$numberDouble", BSON_TYPE_DOUBLE, BSON_JSON_LF_DOUBLE)
+      HANDLE_OPTION (else if, "$symbol", BSON_TYPE_SYMBOL, BSON_JSON_LF_SYMBOL)
+      HANDLE_OPTION (else if,
+                     "$numberDecimal",
+                     BSON_TYPE_DECIMAL128,
+                     BSON_JSON_LF_DECIMAL128)
+      else if (!strcmp ("$timestamp", (const char *) val))
+      {
          bson->bson_type = BSON_TYPE_TIMESTAMP;
          bson->read_state = BSON_JSON_IN_BSON_TYPE_TIMESTAMP_STARTMAP;
-      } else if (!strcmp ("$regularExpression", (const char *) val)) {
+      }
+      else if (!strcmp ("$regularExpression", (const char *) val))
+      {
          bson->bson_type = BSON_TYPE_REGEX;
          bson->read_state = BSON_JSON_IN_BSON_TYPE_REGEX_STARTMAP;
-      } else if (!strcmp ("$dbPointer", (const char *) val)) {
+      }
+      else if (!strcmp ("$dbPointer", (const char *) val))
+      {
          /* start parsing "key": {"$dbPointer": {...}}, save "key" for later */
          _bson_json_buf_set (
             &bson->dbpointer_key, bson->key_buf.buf, bson->key_buf.len);
 
          bson->bson_type = BSON_TYPE_DBPOINTER;
          bson->read_state = BSON_JSON_IN_BSON_TYPE_DBPOINTER_STARTMAP;
-      } else if (!strcmp ("$code", (const char *) val)) {
+      }
+      else if (!strcmp ("$code", (const char *) val))
+      {
          _bson_json_read_code_or_scope_key (
             bson, false /* is_scope */, val, len);
-      } else if (!strcmp ("$scope", (const char *) val)) {
+      }
+      else if (!strcmp ("$scope", (const char *) val))
+      {
          _bson_json_read_code_or_scope_key (
             bson, true /* is_scope */, val, len);
-      } else {
+      }
+      else
+      {
          _bson_json_bad_key_in_type (reader, val);
       }
    } else if (bson->read_state == BSON_JSON_IN_BSON_TYPE_DATE_NUMBERLONG) {
-      if
-         HANDLE_OPTION ("$numberLong", BSON_TYPE_DATE_TIME, BSON_JSON_LF_INT64)
-      else {
+      HANDLE_OPTION (if, "$numberLong", BSON_TYPE_DATE_TIME, BSON_JSON_LF_INT64)
+      else
+      {
          _bson_json_bad_key_in_type (reader, val);
       }
    } else if (bson->read_state == BSON_JSON_IN_BSON_TYPE_TIMESTAMP_VALUES) {
-      if
-         HANDLE_OPTION ("t", BSON_TYPE_TIMESTAMP, BSON_JSON_LF_TIMESTAMP_T)
-      else if
-         HANDLE_OPTION ("i", BSON_TYPE_TIMESTAMP, BSON_JSON_LF_TIMESTAMP_I)
-      else {
+      HANDLE_OPTION (if, "t", BSON_TYPE_TIMESTAMP, BSON_JSON_LF_TIMESTAMP_T)
+      HANDLE_OPTION (
+         else if, "i", BSON_TYPE_TIMESTAMP, BSON_JSON_LF_TIMESTAMP_I)
+      else
+      {
          _bson_json_bad_key_in_type (reader, val);
       }
    } else if (bson->read_state == BSON_JSON_IN_BSON_TYPE_REGEX_VALUES) {
-      if
-         HANDLE_OPTION (
-            "pattern", BSON_TYPE_REGEX, BSON_JSON_LF_REGULAR_EXPRESSION_PATTERN)
-      else if
-         HANDLE_OPTION (
-            "options", BSON_TYPE_REGEX, BSON_JSON_LF_REGULAR_EXPRESSION_OPTIONS)
-      else {
+      HANDLE_OPTION (if,
+                     "pattern",
+                     BSON_TYPE_REGEX,
+                     BSON_JSON_LF_REGULAR_EXPRESSION_PATTERN)
+      HANDLE_OPTION (else if,
+                     "options",
+                     BSON_TYPE_REGEX,
+                     BSON_JSON_LF_REGULAR_EXPRESSION_OPTIONS)
+      else
+      {
          _bson_json_bad_key_in_type (reader, val);
       }
    } else if (bson->read_state == BSON_JSON_IN_BSON_TYPE_BINARY_VALUES) {
-      if
-         HANDLE_OPTION ("base64", BSON_TYPE_BINARY, BSON_JSON_LF_BINARY)
-      else if
-         HANDLE_OPTION ("subType", BSON_TYPE_BINARY, BSON_JSON_LF_TYPE)
-      else {
+      HANDLE_OPTION (if, "base64", BSON_TYPE_BINARY, BSON_JSON_LF_BINARY)
+      HANDLE_OPTION (else if, "subType", BSON_TYPE_BINARY, BSON_JSON_LF_TYPE)
+      else
+      {
          _bson_json_bad_key_in_type (reader, val);
       }
    } else {
       _bson_json_save_map_key (bson, val, len);
-
-      /* in x: {$ref: "collection", $id: {$oid: "..."}, $db: "..." } */
-      if (bson->n > 0) {
-         if (!strcmp ("$ref", (const char *) val)) {
-            STACK_HAS_REF = true;
-            bson->read_state = BSON_JSON_IN_BSON_TYPE;
-            bson->bson_state = BSON_JSON_LF_DBREF;
-         } else if (!strcmp ("$id", (const char *) val)) {
-            STACK_HAS_ID = true;
-         } else if (!strcmp ("$db", (const char *) val)) {
-            bson->read_state = BSON_JSON_IN_BSON_TYPE;
-            bson->bson_state = BSON_JSON_LF_DBREF;
-         }
-      }
    }
 }
 
@@ -1369,11 +1518,6 @@ _bson_json_read_append_regex (bson_json_reader_t *reader,    /* IN */
                                     "Missing \"$regex\" after \"$options\"");
          return;
       }
-      if (!data->regex.has_options) {
-         _bson_json_read_set_error (reader,
-                                    "Missing \"$options\" after \"$regex\"");
-         return;
-      }
    } else if (!data->regex.has_pattern) {
       _bson_json_read_set_error (
          reader, "Missing \"pattern\" after \"options\" in regular expression");
@@ -1720,11 +1864,6 @@ _bson_json_read_end_map (bson_json_reader_t *reader) /* IN */
          bson->read_state = BSON_JSON_IN_BSON_TYPE_DBPOINTER_STARTMAP;
          STACK_POP_DBPOINTER;
       } else {
-         if (STACK_HAS_ID != STACK_HAS_REF) {
-            _bson_json_read_set_error (
-               reader, "%s", "DBRef object must have both $ref and $id keys");
-         }
-
          STACK_POP_DOC (
             bson_append_document_end (STACK_BSON_PARENT, STACK_BSON_CHILD));
       }
@@ -1763,7 +1902,7 @@ _bson_json_read_start_array (bson_json_reader_t *reader) /* IN */
       return;
    }
 
-   if (bson->n < 0) {
+   if (bson->n == -1) {
       STACK_PUSH_ARRAY (_noop ());
    } else {
       _bson_json_read_fixup_key (bson);
@@ -1875,6 +2014,9 @@ _push_callback (jsonsl_t json,
 {
    bson_json_reader_t *reader = (bson_json_reader_t *) json->data;
 
+   BSON_UNUSED (action);
+   BSON_UNUSED (buf);
+
    switch (state->type) {
    case JSONSL_T_STRING:
    case JSONSL_T_HKEY:
@@ -1906,6 +2048,8 @@ _pop_callback (jsonsl_t json,
    double d;
    const char *obj_text;
 
+   BSON_UNUSED (action);
+
    reader = (bson_json_reader_t *) json->data;
    reader_bson = &reader->bson;
 
@@ -1973,6 +2117,8 @@ _error_callback (jsonsl_t json,
 {
    bson_json_reader_t *reader = (bson_json_reader_t *) json->data;
 
+   BSON_UNUSED (state);
+
    if (err == JSONSL_ERROR_CANT_INSERT && *errat == '{') {
       /* start the next document */
       reader->should_reset = true;
@@ -2123,7 +2269,9 @@ bson_json_reader_new (void *data,               /* IN */
    bson_json_reader_t *r;
    bson_json_reader_producer_t *p;
 
-   r = bson_malloc0 (sizeof *r);
+   BSON_UNUSED (allow_multiple);
+
+   r = BSON_ALIGNED_ALLOC0 (bson_json_reader_t);
    r->json = jsonsl_new (STACK_MAX);
    r->json->error_callback = _error_callback;
    r->json->action_callback_PUSH = _push_callback;
