@@ -15,7 +15,7 @@
  */
 
 
-#include <bson.h>
+#include <bson/bson.h>
 
 #include "mongoc-config.h"
 #include "mongoc-counters-private.h"
@@ -35,6 +35,14 @@
 #include "mongoc-crypto-cng-private.h"
 #endif
 
+#ifdef MONGOC_ENABLE_MONGODB_AWS_AUTH
+#include "kms_message/kms_message.h"
+#endif
+
+#ifdef MONGOC_ENABLE_OCSP_OPENSSL
+#include "mongoc-ocsp-cache-private.h"
+#endif
+
 #ifndef MONGOC_NO_AUTOMATIC_GLOBALS
 #pragma message( \
    "Configure the driver with ENABLE_AUTOMATIC_INIT_AND_CLEANUP=OFF.\
@@ -47,10 +55,10 @@
 static void *
 mongoc_cyrus_mutex_alloc (void)
 {
-   mongoc_mutex_t *mutex;
+   bson_mutex_t *mutex;
 
-   mutex = (mongoc_mutex_t *) bson_malloc0 (sizeof (mongoc_mutex_t));
-   mongoc_mutex_init (mutex);
+   mutex = (bson_mutex_t *) bson_malloc0 (sizeof (bson_mutex_t));
+   bson_mutex_init (mutex);
 
    return (void *) mutex;
 }
@@ -59,7 +67,7 @@ mongoc_cyrus_mutex_alloc (void)
 static int
 mongoc_cyrus_mutex_lock (void *mutex)
 {
-   mongoc_mutex_lock ((mongoc_mutex_t *) mutex);
+   bson_mutex_lock ((bson_mutex_t *) mutex);
 
    return SASL_OK;
 }
@@ -68,7 +76,7 @@ mongoc_cyrus_mutex_lock (void *mutex)
 static int
 mongoc_cyrus_mutex_unlock (void *mutex)
 {
-   mongoc_mutex_unlock ((mongoc_mutex_t *) mutex);
+   bson_mutex_unlock ((bson_mutex_t *) mutex);
 
    return SASL_OK;
 }
@@ -77,14 +85,14 @@ mongoc_cyrus_mutex_unlock (void *mutex)
 static void
 mongoc_cyrus_mutex_free (void *mutex)
 {
-   mongoc_mutex_destroy ((mongoc_mutex_t *) mutex);
+   bson_mutex_destroy ((bson_mutex_t *) mutex);
    bson_free (mutex);
 }
 
 #endif /* MONGOC_ENABLE_SASL_CYRUS */
 
 
-static MONGOC_ONCE_FUN (_mongoc_do_init)
+static BSON_ONCE_FUN (_mongoc_do_init)
 {
 #ifdef MONGOC_ENABLE_SASL_CYRUS
    int status;
@@ -131,17 +139,25 @@ static MONGOC_ONCE_FUN (_mongoc_do_init)
 
    _mongoc_handshake_init ();
 
-   MONGOC_ONCE_RETURN;
+#if defined(MONGOC_ENABLE_MONGODB_AWS_AUTH)
+   kms_message_init ();
+#endif
+
+#if defined(MONGOC_ENABLE_OCSP_OPENSSL)
+  _mongoc_ocsp_cache_init ();
+#endif
+
+   BSON_ONCE_RETURN;
 }
 
 void
 mongoc_init (void)
 {
-   static mongoc_once_t once = MONGOC_ONCE_INIT;
-   mongoc_once (&once, _mongoc_do_init);
+   static bson_once_t once = BSON_ONCE_INIT;
+   bson_once (&once, _mongoc_do_init);
 }
 
-static MONGOC_ONCE_FUN (_mongoc_do_cleanup)
+static BSON_ONCE_FUN (_mongoc_do_cleanup)
 {
 #ifdef MONGOC_ENABLE_SSL_OPENSSL
    _mongoc_openssl_cleanup ();
@@ -168,14 +184,22 @@ static MONGOC_ONCE_FUN (_mongoc_do_cleanup)
 
    _mongoc_handshake_cleanup ();
 
-   MONGOC_ONCE_RETURN;
+#if defined(MONGOC_ENABLE_MONGODB_AWS_AUTH)
+   kms_message_cleanup ();
+#endif
+
+#if defined(MONGOC_ENABLE_OCSP_OPENSSL)
+   _mongoc_ocsp_cache_cleanup ();
+#endif
+
+   BSON_ONCE_RETURN;
 }
 
 void
 mongoc_cleanup (void)
 {
-   static mongoc_once_t once = MONGOC_ONCE_INIT;
-   mongoc_once (&once, _mongoc_do_cleanup);
+   static bson_once_t once = BSON_ONCE_INIT;
+   bson_once (&once, _mongoc_do_cleanup);
 }
 
 /*
