@@ -28,6 +28,7 @@
 #include "mongoc-secure-channel-private.h"
 #include "mongoc-stream-tls-secure-channel-private.h"
 #include "mongoc-errno-private.h"
+#include "mongoc-error.h"
 
 
 #undef MONGOC_LOG_DOMAIN
@@ -111,7 +112,7 @@ mongoc_secure_channel_setup_certificate_from_file (const char *filename)
       NULL, /* phCertStore, OUT, HCERTSTORE.., unused, for now */
       NULL, /* phMsg, OUT, HCRYPTMSG, only for PKC7, unused */
       (const void **) &cert /* ppvContext, OUT, the Certificate Context */
-      );
+   );
 
    if (!cert) {
       MONGOC_ERROR ("Failed to extract public key from '%s'. Error 0x%.8X",
@@ -429,7 +430,7 @@ mongoc_secure_channel_setup_crl (
       NULL, /* phCertStore, OUT, HCERTSTORE.., unused, for now */
       NULL, /* phMsg, OUT, HCRYPTMSG, only for PKC7, unused */
       (const void **) &cert /* ppvContext, OUT, the Certificate Context */
-      );
+   );
    bson_free (str);
 
    if (!cert) {
@@ -474,7 +475,9 @@ mongoc_secure_channel_read (mongoc_stream_tls_t *tls,
    ssize_t length;
 
    errno = 0;
-   TRACE ("Wanting to read: %d, timeout is %d", data_length, tls->timeout_msec);
+   TRACE ("Wanting to read: %zu, timeout is %" PRId32,
+          data_length,
+          tls->timeout_msec);
    /* 4th argument is minimum bytes, while the data_length is the
     * size of the buffer. We are totally fine with just one TLS record (few
     *bytes)
@@ -482,7 +485,7 @@ mongoc_secure_channel_read (mongoc_stream_tls_t *tls,
    length = mongoc_stream_read (
       tls->base_stream, data, data_length, 0, tls->timeout_msec);
 
-   TRACE ("Got %d", length);
+   TRACE ("Got %zd", length);
 
    if (length > 0) {
       return length;
@@ -499,10 +502,10 @@ mongoc_secure_channel_write (mongoc_stream_tls_t *tls,
    ssize_t length;
 
    errno = 0;
-   TRACE ("Wanting to write: %d", data_length);
+   TRACE ("Wanting to write: %zd", data_length);
    length = mongoc_stream_write (
       tls->base_stream, (void *) data, data_length, tls->timeout_msec);
-   TRACE ("Wrote: %d", length);
+   TRACE ("Wrote: %zd", length);
 
 
    return length;
@@ -569,9 +572,16 @@ _mongoc_secure_channel_init_sec_buffer_desc (SecBufferDesc *desc,
 }
 
 
+#define MONGOC_LOG_AND_SET_ERROR(ERROR, DOMAIN, CODE, ...) \
+   do {                                                    \
+      MONGOC_ERROR (__VA_ARGS__);                          \
+      bson_set_error (ERROR, DOMAIN, CODE, __VA_ARGS__);   \
+   } while (0)
+
 bool
 mongoc_secure_channel_handshake_step_1 (mongoc_stream_tls_t *tls,
-                                        char *hostname)
+                                        char *hostname,
+                                        bson_error_t *error)
 {
    SecBuffer outbuf;
    ssize_t written = -1;
@@ -609,11 +619,14 @@ mongoc_secure_channel_handshake_step_1 (mongoc_stream_tls_t *tls,
       &outbuf_desc,                       /* pOutput OUT param */
       &secure_channel->ret_flags,         /* pfContextAttr OUT param */
       &secure_channel->ctxt->time_stamp   /* ptsExpiry OUT param */
-      );
+   );
 
    if (sspi_status != SEC_I_CONTINUE_NEEDED) {
-      MONGOC_ERROR ("initial InitializeSecurityContext failed: %ld",
-                    sspi_status);
+      MONGOC_LOG_AND_SET_ERROR (error,
+                                MONGOC_ERROR_STREAM,
+                                MONGOC_ERROR_STREAM_SOCKET,
+                                "initial InitializeSecurityContext failed: %ld",
+                                sspi_status);
       return false;
    }
 
@@ -626,10 +639,13 @@ mongoc_secure_channel_handshake_step_1 (mongoc_stream_tls_t *tls,
    FreeContextBuffer (outbuf.pvBuffer);
 
    if (outbuf.cbBuffer != (size_t) written) {
-      MONGOC_ERROR ("failed to send initial handshake data: "
-                    "sent %zd of %lu bytes",
-                    written,
-                    outbuf.cbBuffer);
+      MONGOC_LOG_AND_SET_ERROR (error,
+                                MONGOC_ERROR_STREAM,
+                                MONGOC_ERROR_STREAM_SOCKET,
+                                "failed to send initial handshake data: "
+                                "sent %zd of %lu bytes",
+                                written,
+                                outbuf.cbBuffer);
       return false;
    }
 
@@ -647,7 +663,8 @@ mongoc_secure_channel_handshake_step_1 (mongoc_stream_tls_t *tls,
 
 bool
 mongoc_secure_channel_handshake_step_2 (mongoc_stream_tls_t *tls,
-                                        char *hostname)
+                                        char *hostname,
+                                        bson_error_t *error)
 {
    mongoc_stream_tls_secure_channel_t *secure_channel =
       (mongoc_stream_tls_secure_channel_t *) tls->ctx;
@@ -666,6 +683,12 @@ mongoc_secure_channel_handshake_step_2 (mongoc_stream_tls_t *tls,
    TRACE ("%s", "SSL/TLS connection with endpoint (step 2/3)");
 
    if (!secure_channel->cred || !secure_channel->ctxt) {
+      MONGOC_LOG_AND_SET_ERROR (
+         error,
+         MONGOC_ERROR_STREAM,
+         MONGOC_ERROR_STREAM_SOCKET,
+         "required TLS credentials or context not provided");
+
       return false;
    }
 
@@ -695,8 +718,12 @@ mongoc_secure_channel_handshake_step_2 (mongoc_stream_tls_t *tls,
                return true;
             }
 
-            MONGOC_ERROR (
+            MONGOC_LOG_AND_SET_ERROR (
+               error,
+               MONGOC_ERROR_STREAM,
+               MONGOC_ERROR_STREAM_SOCKET,
                "failed to receive handshake, SSL/TLS connection failed");
+
             return false;
          }
 
@@ -729,7 +756,10 @@ mongoc_secure_channel_handshake_step_2 (mongoc_stream_tls_t *tls,
       _mongoc_secure_channel_init_sec_buffer_desc (&outbuf_desc, outbuf, 3);
 
       if (inbuf[0].pvBuffer == NULL) {
-         MONGOC_ERROR ("unable to allocate memory");
+         MONGOC_LOG_AND_SET_ERROR (error,
+                                   MONGOC_ERROR_STREAM,
+                                   MONGOC_ERROR_STREAM_SOCKET,
+                                   "unable to allocate memory");
          return false;
       }
 
@@ -789,10 +819,14 @@ mongoc_secure_channel_handshake_step_2 (mongoc_stream_tls_t *tls,
                   tls, outbuf[i].pvBuffer, outbuf[i].cbBuffer);
 
                if (outbuf[i].cbBuffer != (size_t) written) {
-                  MONGOC_ERROR ("failed to send next handshake data: "
-                                "sent %zd of %lu bytes",
-                                written,
-                                outbuf[i].cbBuffer);
+                  MONGOC_LOG_AND_SET_ERROR (
+                     error,
+                     MONGOC_ERROR_STREAM,
+                     MONGOC_ERROR_STREAM_SOCKET,
+                     "failed to send next handshake data: "
+                     "sent %zd of %lu bytes",
+                     written,
+                     outbuf[i].cbBuffer);
                   return false;
                }
             }
@@ -805,22 +839,38 @@ mongoc_secure_channel_handshake_step_2 (mongoc_stream_tls_t *tls,
       } else {
          switch (sspi_status) {
          case SEC_E_WRONG_PRINCIPAL:
-            MONGOC_ERROR ("SSL Certification verification failed: hostname "
-                          "doesn't match certificate");
+            MONGOC_LOG_AND_SET_ERROR (
+               error,
+               MONGOC_ERROR_STREAM,
+               MONGOC_ERROR_STREAM_SOCKET,
+               "SSL Certification verification failed: hostname "
+               "doesn't match certificate");
             break;
 
          case SEC_E_UNTRUSTED_ROOT:
-            MONGOC_ERROR ("SSL Certification verification failed: Untrusted "
-                          "root certificate");
+            MONGOC_LOG_AND_SET_ERROR (
+               error,
+               MONGOC_ERROR_STREAM,
+               MONGOC_ERROR_STREAM_SOCKET,
+               "SSL Certification verification failed: Untrusted "
+               "root certificate");
             break;
 
          case SEC_E_CERT_EXPIRED:
-            MONGOC_ERROR ("SSL Certification verification failed: certificate "
-                          "has expired");
+            MONGOC_LOG_AND_SET_ERROR (
+               error,
+               MONGOC_ERROR_STREAM,
+               MONGOC_ERROR_STREAM_SOCKET,
+               "SSL Certification verification failed: certificate "
+               "has expired");
             break;
          case CRYPT_E_NO_REVOCATION_CHECK:
-            MONGOC_ERROR ("SSL Certification verification failed: certificate "
-                          "does not include revocation check.");
+            MONGOC_LOG_AND_SET_ERROR (
+               error,
+               MONGOC_ERROR_STREAM,
+               MONGOC_ERROR_STREAM_SOCKET,
+               "SSL Certification verification failed: certificate "
+               "does not include revocation check.");
             break;
 
          case SEC_E_INSUFFICIENT_MEMORY:
@@ -850,11 +900,15 @@ mongoc_secure_channel_handshake_step_2 (mongoc_stream_tls_t *tls,
                            (LPTSTR) &msg,
                            0,
                            NULL);
-            MONGOC_ERROR ("Failed to initialize security context, error code: "
-                          "0x%04X%04X: %s",
-                          (unsigned int) (sspi_status >> 16) & 0xffff,
-                          (unsigned int) sspi_status & 0xffff,
-                          msg);
+            MONGOC_LOG_AND_SET_ERROR (
+               error,
+               MONGOC_ERROR_STREAM,
+               MONGOC_ERROR_STREAM_SOCKET,
+               "Failed to initialize security context, error code: "
+               "0x%04X%04X: %s",
+               (unsigned int) (sspi_status >> 16) & 0xffff,
+               (unsigned int) sspi_status & 0xffff,
+               msg);
             LocalFree (msg);
          }
          }
@@ -914,7 +968,8 @@ mongoc_secure_channel_handshake_step_2 (mongoc_stream_tls_t *tls,
 
 bool
 mongoc_secure_channel_handshake_step_3 (mongoc_stream_tls_t *tls,
-                                        char *hostname)
+                                        char *hostname,
+                                        bson_error_t *error)
 {
    mongoc_stream_tls_secure_channel_t *secure_channel =
       (mongoc_stream_tls_secure_channel_t *) tls->ctx;
@@ -924,12 +979,19 @@ mongoc_secure_channel_handshake_step_3 (mongoc_stream_tls_t *tls,
    TRACE ("SSL/TLS connection with %s (step 3/3)", hostname);
 
    if (!secure_channel->cred) {
+      MONGOC_LOG_AND_SET_ERROR (error,
+                                MONGOC_ERROR_STREAM,
+                                MONGOC_ERROR_STREAM_SOCKET,
+                                "required TLS credentials not provided");
       return false;
    }
 
    /* check if the required context attributes are met */
    if (secure_channel->ret_flags != secure_channel->req_flags) {
-      MONGOC_ERROR ("Failed handshake");
+      MONGOC_LOG_AND_SET_ERROR (error,
+                                MONGOC_ERROR_STREAM,
+                                MONGOC_ERROR_STREAM_SOCKET,
+                                "Failed handshake");
 
       return false;
    }
